# CMSIS Pack Manager
# Copyright (c) 2017-2021 Arm Limited
# Copyright (c) 2021 Chris Reed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time
import collections
from os.path import join, dirname, exists
from shutil import rmtree
from json import load
from zipfile import ZipFile
from appdirs import user_data_dir
from .cmsis_pack_manager import ffi, lib


class _RaiseRust(object):
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exe_tb):
        maybe_err = ffi.gc(lib.err_get_last_message(),
                           lib.err_last_message_free)
        if maybe_err:
            raise Exception(ffi.string(maybe_err))


class CmsisPackRef(collections.namedtuple(
        "CmsisPackRef",
        "vendor, pack, version"
)):
    def get_pdsc_name(self):
        return "{}.{}.{}.pdsc".format(self.vendor, self.pack, self.version)

    def get_pack_name(self):
        return join(self.vendor, self.pack, "{}.pack".format(self.version))

    def __str__(self):
        return f"{self.vendor}.{self.pack}.{self.version}"


class Cache (object):
    """ The Cache object is the only relevant API object at the moment

    Constructing the Cache object does not imply any caching.
    A user of the API must explicitly call caching functions.

    :param silent: A boolean that, when True, significantly reduces the
                   printing of this Object
    :type silent: bool
    :param no_timeouts: A boolean that, when True, disables the default
                        connection timeout and low speed timeout for
                        downloading things.
    :type no_timeouts: bool
    """
    def __init__(
            self,
            silent,
            __,
            json_path=None,
            data_path=None,
            vidx_list=None
    ):
        default_path = user_data_dir('cmsis-pack-manager')
        json_path = default_path if not json_path else json_path
        self._index = {}
        self._aliases = {}
        self.index_path = join(json_path, "index.json")
        self.aliases_path = join(json_path, "aliases.json")
        self.data_path = default_path if not data_path else data_path
        self.vidx_list = vidx_list
        self.silent = silent

    def get_flash_algorithm_binary(self, device_name, all=False):
        """Retrieve the flash algorithm file for a particular part.

        Assumes that both the PDSC and the PACK file associated with that part
        are in the cache.

        :param device_name: The exact name of a device
        :param all: Return an iterator of all flash algos for this device
        :type device_name: str
        :return: A file-like object that, when read, is the ELF file that
                 describes the flashing algorithm
        :return: A file-like object that, when read, is the ELF file that
                 describes the flashing algorithm.  When "all" is set to
                 True then an iterator for file-like objects is returned
        :rtype: ZipExtFile or ZipExtFile iterator if all is True
        """
        device = self.index[device_name]
        pack = self.pack_from_cache(device)
        algo_itr = (pack.open(algo['file_name'].replace(u'\\', '/')) for algo
                    in device['algorithms'])
        return algo_itr if all else algo_itr.next()

    @property
    def index(self):
        """An index of most of the important data in all cached PDSC files.

        :Example:

        >>> from ArmPackManager import Cache
        >>> a = Cache()
        >>> a.index["LPC1768"]
        {u'algorithms': [{u'default': False,
                          u'file_name': u'Flash/LPC_IAP_512.FLM',
                          u'size': 524288,
                          u'start': 0}],
         u'from_pack': {u'pack': u'LPC1700_DFP',
                        u'vendor': u'Keil',
                        u'version': u'2.4.0'},
         u'memories': {u'IRAM1': {u'access': {u'execute': False,
                                              u'read': True,
                                              u'write': True},
                                  u'size': 32768,
                                  u'start': 268435456,
                                  u'startup': False},
                       u'IRAM2': {u'access': {u'execute': False,
                                              u'read': True,
                                              u'write': True},
                                  u'size': 32768,
                                  u'start': 537378816,
                                  u'startup': False},
                       u'IROM1': {u'access': {u'execute': True,
                                              u'read': True,
                                              u'write': False},
                                  u'size': 524288,
                                  u'start': 0,
                                  u'startup': False}},
         u'name': u'LPC1768'}

        """
        if not self._index:
            try:
                with open(self.index_path) as i:
                    self._index = load(i)
            except IOError:
                return {}
        return self._index

    @property
    def aliases(self):
        """An index of the boards in all CMSIS Pack Descriptions.

        :Example:

        >>> from cmsis_pack_manager import Cache
        >>> a = Cache()
        >>> a.aliases["LPC1788-32 Developers Kit"]
        {"name": "LPC1788-32 Developers Kit",
         "mounted_devices": ["LPC1788"]}
        """

        if not self._aliases:
            try:
                with open(self.aliases_path) as i:
                    self._aliases = load(i)
            except IOError:
                return {}
        return self._aliases

    def cache_everything(self):
        """Cache every CMSIS Pack and generate an index.

        .. note:: This process may use 5GB of drive space and take upwards of
        2 minutes to complete.
        """
        parsed_packs = self.cache_descriptors()
        progress_fn = None if self.silent else self._verbose_on_tick_fn
        self._call_rust_update_packs(parsed_packs, progress_fn)

    def packs_for_devices(self, devices):
        """Create a list of all CMSIS Packs required to support the
        list of devices

        :return:
        A list of CmsisPackRef objects
        """
        to_ret = {}
        for dev in devices:
            pack_ref = CmsisPackRef(
                dev["from_pack"]["vendor"],
                dev["from_pack"]["pack"],
                dev["from_pack"]["version"],
            )
            to_ret[pack_ref] = True
        return list(to_ret.keys())

    def download_pack_list(self, pack_list):
        pdsc_index = ffi.gc(
            lib.update_pdsc_index_new(),
            lib.update_pdsc_index_free
        )
        for pack_ref in pack_list:
            pdsc_path = join(self.data_path, pack_ref.get_pdsc_name())
            cpdsc_path = ffi.new("char[]", pdsc_path.encode("utf-8"))
            with _RaiseRust():
                lib.update_pdsc_index_push(pdsc_index, cpdsc_path)
        parsed_packs = self._call_rust_parse(pdsc_index)
        progress_fn = None if self.silent else self._verbose_on_tick_fn
        self._call_rust_update_packs(parsed_packs, progress_fn)

    def _call_rust_update_packs(self, parsed_packs, on_tick_fn):
        if self.data_path:
            cdata_path = ffi.new("char[]", self.data_path.encode("utf-8"))
        else:
            cdata_path = ffi.NULL
        with _RaiseRust():
            poll_obj = lib.update_packs(cdata_path, parsed_packs)
        return self._poll_rust_update(poll_obj, on_tick_fn)

    def _verbose_on_tick_fn(self, total, current):
        if total:
            total = "{:03}".format(total)
        else:
            total = "???"
        sys.stdout.write(
            "Downloading descriptors ({:03}/{})\r".format(current, total)
        )
        sys.stdout.flush()

    def _call_rust_update(self, on_tick_fn):
        if self.data_path:
            cdata_path = ffi.new("char[]", self.data_path.encode("utf-8"))
        else:
            cdata_path = ffi.NULL
        if self.vidx_list:
            cvidx_path = ffi.new("char[]", self.vidx_list.encode("utf-8"))
        else:
            cvidx_path = ffi.NULL
        with _RaiseRust():
            poll_obj = lib.update_pdsc_index(cdata_path, cvidx_path)
        return self._poll_rust_update(poll_obj, on_tick_fn)

    def _poll_rust_update(self, poll_obj, on_tick_fn):
        with _RaiseRust():
            total_downloads = None
            current_downloads = 0
            while not lib.update_pdsc_poll(poll_obj):
                prev_downloads = current_downloads
                time.sleep(1/2)
                message = ffi.gc(
                    lib.update_pdsc_get_status(poll_obj),
                    lib.update_pdsc_status_free
                )
                while message:
                    if message.is_size:
                        total_downloads = message.size
                    else:
                        current_downloads += 1
                    message = ffi.gc(
                        lib.update_pdsc_get_status(poll_obj),
                        lib.update_pdsc_status_free
                    )
                if on_tick_fn and current_downloads != prev_downloads:
                    on_tick_fn(total_downloads, current_downloads)
            pdsc_index = ffi.gc(
                lib.update_pdsc_result(poll_obj), lib.update_pdsc_index_free
            )
        return pdsc_index

    def _call_rust_parse(self, pdsc_index):
        with _RaiseRust():
            parsed_packs = ffi.gc(lib.parse_packs(pdsc_index),
                                  lib.parse_packs_free)
        return parsed_packs

    def _call_rust_dump(self, parsed_packs):
        if self.index_path:
            cindex_path = ffi.new("char[]", self.index_path.encode("utf-8"))
        else:
            cindex_path = ffi.NULL
        if self.aliases_path:
            calias_path = ffi.new("char[]", self.aliases_path.encode("utf-8"))
        else:
            calias_path = ffi.NULL
        with _RaiseRust():
            lib.dump_pdsc_json(parsed_packs, cindex_path, calias_path)
        return parsed_packs

    def cache_descriptors(self):
        """Cache all Pack Descriptions and generate an index of them.

        .. note:: This process may use 14MB of drive space and take upwards of
        10 seconds.
        """
        progress_fn = None if self.silent else self._verbose_on_tick_fn
        pdsc_index = self._call_rust_update(progress_fn)
        parsed_packs = self._call_rust_parse(pdsc_index)
        self._call_rust_dump(parsed_packs)
        return parsed_packs

    def cache_clean(self):
        """Clean the entire cache."""
        if exists(self.data_path):
            rmtree(self.data_path)
        json_path = dirname(self.index_path)
        if exists(json_path):
            rmtree(json_path)

    def pdsc_from_cache(self, device):
        """Low level inteface for extracting a PDSC file from the cache.

        Assumes that the file specified is a PDSC file and is in the cache.

        :param url: The URL of a PDSC file.
        :type url: str
        :return: A parsed representation of the PDSC file.
        :rtype: Open file
        """
        from_pack = device['from_pack']
        dest = join(self.data_path, "{}.{}.{}.pdsc".format(
            from_pack['vendor'], from_pack['pack'], from_pack['version']))
        return open(dest, "r")

    def pack_from_cache(self, device):
        """Low level inteface for extracting a PACK file from the cache.

        Assumes that the file specified is a PACK file and is in the cache.

        :param url: The URL of a PACK file.
        :type url: str
        :return: A parsed representation of the PACK file.
        :rtype: ZipFile
        """
        from_pack = device['from_pack']
        return ZipFile(join(self.data_path,
                            from_pack['vendor'],
                            from_pack['pack'],
                            from_pack['version'] + ".pack"))

    @staticmethod
    def find_pdsc(zipfile):
        """Find the PDSC file within a PACK file

        :param zipfile: The PACK to scan
        :type zipfile: ZipFile
        :return: The location of the PDSC file within the PACK file
        :rtype: str
        """
        for zipinfo in zipfile.infolist():
            if (zipinfo.filename.upper().endswith(".PDSC")):
                return zipinfo.filename
        return None

    def add_pack_from_path(self, path):
        if path:
            cpack_path = ffi.new("char[]", path.encode("utf-8"))
        else:
            cpack_path = ffi.NULL
        with _RaiseRust():
            pack_files = ffi.gc(lib.pack_from_path(cpack_path),
                                lib.update_pdsc_index_free)
        parsed_packs = self._call_rust_parse(pack_files)
        self._call_rust_dump(parsed_packs)
        return parsed_packs



# ARM Pack Manager
# Copyright (c) 2017-2021 Arm Limited
# Copyright (c) 2021 Chris Reed
# Copyright (c) 2021 Noah Pendleton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function, absolute_import

import argparse
import operator
import sys
from os.path import join, dirname, exists
from os import makedirs
from functools import reduce
from json import dump
import yaml
from cmsis_pack_manager import Cache

parser = argparse.ArgumentParser(
    description='A utility that keeps your cache of pack files up to date.'
)
subparsers = parser.add_subparsers(title="Commands")


def subcommand(name, *args, **kwargs):
    def subcommand(command):
        subparser = subparsers.add_parser(name, **kwargs)

        for arg in args:
            arg = dict(arg)
            opt = arg['name']
            del arg['name']

            if isinstance(opt, str):
                subparser.add_argument(opt, **arg)
            else:
                subparser.add_argument(*opt, **arg)

        subparser.add_argument(
            "-v", "--verbose",
            action="store_true",
            dest="verbose",
            help="Verbose diagnostic output",
        )
        subparser.add_argument(
            "--and",
            action="store_true",
            dest="intersection",
            help="Combine search terms as if with an `and`",
        )
        subparser.add_argument(
            "--or",
            action="store_false",
            dest="intersection",
            help="Combine search terms as if with an `or`",
        )
        subparser.add_argument(
            "--union",
            action="store_false",
            dest="intersection",
            help="Combine search terms as if with a set union",
        )
        subparser.add_argument(
            "--intersection",
            action="store_true",
            dest="intersection",
            help="Combine search terms as if with a set intersection",
        )
        subparser.add_argument("--vidx-list", dest="vidx_list")
        subparser.add_argument("--data-path", dest="data_path")
        subparser.add_argument("--json-path", dest="json_path")

        def thunk(parsed_args):
            cache = Cache(not parsed_args.verbose, False,
                          vidx_list=parsed_args.vidx_list,
                          data_path=parsed_args.data_path,
                          json_path=parsed_args.json_path)
            argv = [
                arg['dest'] if 'dest' in arg else arg['name'] for arg in args
            ]
            argv = [(arg if isinstance(arg, str) else arg[-1]).strip('-')
                    for arg in argv]
            argv = {arg: vars(parsed_args)[arg] for arg in argv
                    if vars(parsed_args)[arg] is not None}

            return command(cache, **argv)

        subparser.set_defaults(command=thunk)
        return command
    return subcommand


def fuzzy_find(matches, options, oper=operator.and_):
    return reduce(oper, (set(filter(lambda x: match in x, options))
                         for match in matches))


CACHE_CHOICES = [
    "descriptors", "desc", "pdsc",
    "everything", "every", "packs",
    "devices", "devs",
    "clean", "rm"
]

CACHE_CHOICES_HELP = (
    "Caching action; "
    "descriptors, desc, pdsc: download only pack descriptors; "
    "everything, every, packs: download packs and descriptors; "
    "devices, devs: download the packs for the listed devices; "
    "clean, rm: remove downloaded packs and descriptors"
)


@subcommand(
    'cache',
    dict(name="action", choices=CACHE_CHOICES, help=CACHE_CHOICES_HELP),
    dict(name="devices", nargs="*"),
    help="Download or remove pack and pack descriptions in the cache"
)
def command_cache(cache, action, devices, verbose=False, intersection=True):
    if action in ("everything", "every", "packs"):
        print("Caching all packs; this will take some time. Press Control-C to quit.")
        cache.cache_everything()
        print("Packs cached")
        return True
    elif action in ("descriptors", "desc", "pdsc"):
        print("Caching descriptors")
        cache.cache_descriptors()
        print("Descriptors cached")
        return True
    elif action in ("clean", "rm"):
        cache.cache_clean()
        print("Cache cleaned")
        return True
    elif action in ("devices", "devs"):
        print("Caching descriptors")
        cache.cache_descriptors()
        devices = [cache.index[dev] for dev in devices]
        packs = cache.packs_for_devices(devices)
        print("Downloading packs: ")
        for pack in packs:
            print("  {}".format(pack))
        print("Press Control-C to quit.")
        cache.download_pack_list(packs)
    else:
        print("No action specified; nothing to do")


@subcommand('find-part',
            dict(name='matches', nargs="+",
                 help="Words to match to processors"),
            dict(name=['-l', "--long"], action="store_true",
                 help="Print out part details with part"),
            dict(name=['-p', '--parts-only'], action="store_false",
                 dest="print_aliases"),
            dict(name=['-a', '--aliases-only'], action="store_false",
                 dest="print_parts"),
            help="Find a part and its description within the cache")
def command_find_part(cache, matches, long=False, intersection=True,
                      print_aliases=True, print_parts=True):
    op = operator.and_ if intersection else operator.or_
    to_dump = {} if long else []
    if print_parts:
        for part in fuzzy_find(matches, cache.index.keys(), op):
            if long:
                to_dump.update({part: cache.index[part]})
            else:
                to_dump.append(part)
    if print_aliases:
        for alias in fuzzy_find(matches, cache.aliases.keys(), op):
            if long:
                if cache.aliases[alias]["mounted_devices"]:
                    part = cache.aliases[alias]["mounted_devices"][0]
                    try:
                        to_dump.update({alias: cache.index[part]})
                    except KeyError:
                        to_dump.update({
                            alias: "Could not find part: %s" % part
                        })
            else:
                to_dump.append(alias)
    print(yaml.safe_dump(to_dump, default_flow_style=None if long else False))


@subcommand('dump-parts',
            dict(name='out', help='Directory to dump to'),
            dict(name='parts', nargs='+', help='Parts to dump'),
            help='Create a directory with an `index.json` describing the part '
            'and all of the associated flashing algorithms.')
def command_dump_parts(cache, out, parts, intersection=False):
    op = operator.and_ if intersection else operator.or_
    index = {part: cache.index[part] for part
             in fuzzy_find(parts, cache.index, op)}
    if not exists(out):
        makedirs(out)
    for n, p in index.items():
        try:
            for algo in p['algorithms']:
                if not exists(join(out, dirname(algo['file_name']))):
                    makedirs(join(out, dirname(algo['file_name'])))
                with open(join(out, algo['file_name']), "wb") as fd:
                    fd.write(cache.pack_from_cache(p)
                             .open(algo['file_name'])
                             .read())
        except KeyError:
            print("[Warning] {} does not have an associated flashing algorithm"
                  .format(n))
    with open(join(out, "index.json"), "w") as fd:
        dump(index, fd)


@subcommand('add-packs',
            dict(name='path', nargs='+',
                 help='path to pdsc to add into the index'),
            help='add contents of pdsc files into the index')
def command_add_packs(cache, path, intersection=False):
    for p in path:
        cache.add_pack_from_path(p)


@subcommand('print-cache-dir',
            help="Print out the cache directory")
def command_print_cache_dir(cache):
    print(cache.data_path)

def get_argparse():
    return parser


def main():
    if (len(sys.argv) < 2):
        parser.print_help()
    else:
        args = parser.parse_args()
        args.command(args)
